;; This part of the configuration only deals with features that are built-in to
;; Emacs.  The aim is to use it as the init file when using Emacs as the
;; Sudo editor or Git editor.  Also, it must be usable with Emacs 28, hence it
;; contains calls to `global-set-key' and `define-key'.

;;;; General Cleanup
(customize-set-variable 'inhibit-startup-message t)
(customize-set-variable 'initial-scratch-message "")

(tooltip-mode -1)
(menu-bar-mode -1)
(save-place-mode 1)

(dolist (mode-func '(tool-bar-mode scroll-bar-mode))
  ;; this is sketchy but prevents errors when some modes aren't available
  (when (fboundp mode-func)
    (funcall mode-func -1)))

(setq-default fill-column 80)

;;;; Make the *scratch* Buffer be text-mode
(customize-set-variable 'initial-major-mode 'text-mode)

;;;; Do not Ask for Killing Buffers
(setq kill-buffer-query-functions nil)

;;;; Parentheses
(electric-pair-mode 1)
(show-paren-mode 1)

;;;; Useful Identity Variables
(customize-set-variable 'user-full-name "Tristan Riehs")
(customize-set-variable 'user-mail-address "tristan.riehs@bordeaux-inp.fr")

;;;; Line and Column Numbers in the Mode Line
(line-number-mode 1)
(column-number-mode 1)

;;;; Line Numbers on the Left in `prog-mode'
(add-hook 'prog-mode-hook 'display-line-numbers-mode)

;;;; Edit .cfg Files in `conf-mode'
(add-to-list 'auto-mode-alist '("\\.cfg?" . conf-mode))

;;;; Better Support for universal-argument on an AZERTY Layout
;; IDEA Make this a package
(defun tr/add-digit-to-prefix-arg (digit)
  "Add DIGIT to the current prefix argument."
  (prefix-command-preserve-state)
  (setq prefix-arg (cond ((integerp current-prefix-arg)
			  (+ (* current-prefix-arg 10)
			     (if (< current-prefix-arg 0) (- digit) digit)))
			 ((eq current-prefix-arg '-)
			  ;; Treat -0 as just -, so that -01 will work.
			  (if (zerop digit) '- (- digit)))
			 (t
			  digit)))
  (universal-argument--mode))

(defun tr/add-digit-to-prefix-arg-func (digit)
  "Build the corresponding command.

The command takes no argument, it may be used for mapping to
a key in `universal-argument-map'."
  (interactive)
  `(lambda ()
     "Built by `add-digit-to-prefix-arg-func'."
     (interactive)
     (prefix-command-preserve-state)
     (add-digit-to-prefix-arg ,digit)
     (universal-argument--mode)))

(dolist (pair '(("&"	. 1)
		("é"	. 2)
		("\""	. 3)
		("'"	. 4)
		("("	. 5)
		("-"	. 6) ;; prevents from using negative arg, but I do not mind
		("è"	. 7)
		("_"	. 8)
		("ç"	. 9)
		("à"	. 0)))
  (define-key universal-argument-map (kbd (car pair))
	      (tr/add-digit-to-prefix-arg-func (cdr pair))))

;;;; More Convenient Window Key Bindings for AZERTY
(global-set-key (kbd "C-x &")  'delete-other-windows)
(global-set-key (kbd "C-x é")  'split-window-below)
(global-set-key (kbd "C-x \"") 'split-window-right)
(global-set-key (kbd "C-x à")  'delete-window)

;;;; Dired
(with-eval-after-load 'dired
  (customize-set-variable 'dired-listing-switches
			  "-Agho --group-directories-first")
  (customize-set-variable 'dired-kill-when-opening-new-dired-buffer t)
  (define-key dired-mode-map (kbd "C-<return>") 'dired-up-directory))

;;;; Backups
(customize-set-variable 'backup-directory-alist
			(list (cons "." (expand-file-name "backups"
							  user-emacs-directory))))

;;;; Completion
(customize-set-variable 'tab-always-indent t)
(global-set-key (kbd "M-<tab>") 'completion-at-point)
(customize-set-variable 'read-file-name-completion-ignore-case t)
(customize-set-variable 'read-buffer-completion-ignore-case t)
(customize-set-variable 'completion-ignore-case t)

;;;; Utils
(defun tr/kill-forward-char (n)
  "Like `delete-forward-char', but always kill."
  (interactive "P")
  (delete-forward-char (or n 1) t))

(global-set-key (kbd "C-d") 'tr/kill-forward-char)
(with-eval-after-load 'cc-mode
  (define-key c-mode-map (kbd "C-d") 'tr/kill-forward-char))

(defun tr/insert-slanted-double-quote ()
  "Insert “ and ” at point."
  (interactive)
  (insert "“”")
  (forward-char -1))

(global-set-key (kbd "C-c \"") 'tr/insert-slanted-double-quote)

(defun tr/newline-before-word ()
  "Insert a newline before the word at point.

The point is then moved back at the end of the word."
  (interactive)
  (backward-word)
  (backward-delete-char 1) ; delete the whitespace
  (insert "\n")
  (forward-word))

(global-set-key (kbd "C-<backspace>") 'tr/newline-before-word)

(defun crux-eval-and-replace ()
  "Replace the preceding sexp with its value.

Taken from :
https://emacs.stackexchange.com/a/47898/40641
https://github.com/bbatsov/crux"
  (interactive)
  (let ((value (eval (elisp--preceding-sexp))))
    (backward-kill-sexp)
    (insert (format "%S" value))))

(global-set-key (kbd "C-c e") 'crux-eval-and-replace)

(global-set-key (kbd "C-c w") 'whitespace-mode)
(global-set-key (kbd "C-c f") 'fixup-whitespace)
(global-set-key (kbd "C-c s") 'sort-lines)

(define-key indent-rigidly-map (kbd "C-f") 'indent-rigidly-right)
(define-key indent-rigidly-map (kbd "C-b") 'indent-rigidly-left)

;;;; Enable some Commands that are Disabled by Default
(put 'upcase-region 'disabled nil)
(put 'capitalize-region 'disabled nil)
(put 'downcase-region 'disabled nil)

;;;; C
(with-eval-after-load 'cc-mode
  (modify-syntax-entry ?_ "w" c-mode-syntax-table))
(setq-default c-default-style "linux")
(setq-default c-indent-level 8)
(setq-default c-indent-tabs-mode t)
(setq-default tab-width 8)

;;;; Help Tweaks
(defun tr/run-and-switch-buffer (command)
  "Run COMMAND and `other-window'."
  (call-interactively command)
  (call-interactively 'other-window))

(defun tr/describe-function ()
  "Run `describe-function' and `other-window'."
  (interactive)
  (tr/run-and-switch-buffer 'describe-function))

(global-set-key (kbd "C-h f") 'tr/describe-function)

(defun tr/describe-variable ()
  "Run `describe-variable' and `other-window'."
  (interactive)
  (tr/run-and-switch-buffer 'describe-variable))

(global-set-key (kbd "C-h v") 'tr/describe-variable)

(defun tr/describe-key ()
  "Run `describe-key' and `other-window'."
  (interactive)
  (tr/run-and-switch-buffer 'describe-key))

(global-set-key (kbd "C-h k") 'tr/describe-key)

(defun tr/describe-mode ()
  "Run `describe-mode' and `other-window'."
  (interactive)
  (tr/run-and-switch-buffer 'describe-mode))

(global-set-key (kbd "C-h m") 'tr/describe-mode)

(defun tr/describe-package ()
  "Run `describe-package' and `other-window'."
  (interactive)
  (tr/run-and-switch-buffer 'describe-package))

(global-set-key (kbd "C-h P") 'tr/describe-package)

;;;; Custom File
(customize-set-variable 'custom-file
			(expand-file-name "custom.el" user-emacs-directory))

(when (file-exists-p custom-file)
  (load custom-file))
