;;;; Load Core Config
(let ((core-config-file (concat (getenv "DOTFILES_DIR") "/emacs/core.el")))
  (when (file-exists-p core-config-file)
    (load core-config-file)))

;;;; Dismiss Intructions
(with-eval-after-load 'server
  (customize-set-variable 'server-client-instructions nil))

;;;; Utils
(global-set-key (kbd "C-x K") 'kill-current-buffer)

(defun tr/expand-from-dotfiles-dir (file)
  "Calls `expand-file-name' from the dotfiles directory."
  (expand-file-name file (getenv "DOTFILES_DIR")))

;;;; Package Archives
(require 'package)
(package-initialize)
(add-to-list 'package-archives '("melpa-stable" . "https://stable.melpa.org/packages/") t)
(add-to-list 'package-archives '("melpa" . "https://melpa.org/packages/") t)
;; (package-refresh-contents)

;;;; Defer Package Loading by Default
(customize-set-variable 'use-package-always-defer t)

;;;; Load Other Config Parts
(dolist (config '("org" "dired" "dev"))
  (load (tr/expand-from-dotfiles-dir (concat "emacs/" config ".el"))))

;;;; Built-in Packages
(with-eval-after-load 'elisp-mode
  (modify-syntax-entry ?- "w" emacs-lisp-mode-syntax-table)
  (add-hook 'emacs-lisp-mode-hook
	    (lambda ()
	      (push '("lambda" . "λ") prettify-symbols-alist)
	      (prettify-symbols-mode 1))))

(use-package which-key
  :defer nil
  :custom
  (which-key-idle-delay 0.8)
  :config
  (which-key-mode 1))

(use-package dabbrev
  :defer nil)

(use-package savehist
  :defer nil
  :config
  (savehist-mode))

(use-package recentf
  :defer nil
  :config
  (recentf-mode 1))

(use-package cc-mode
  :config
  (defun tr/c-electric-lt ()
    "Electrically insert a \"<\" and a \">\".

The \">\" character is only inserted in a C preprocessor
instruction, if the character at the beginning of the
current line is \"#\"."
    (interactive)
    (when electric-pair-mode
      (if (save-excursion
	    (beginning-of-line)
	    (equal (char-after) ?#))
	  (progn
	    (insert "<>")
	    (forward-char -1))
	(insert ?<))))

  (defun tr/c-electric-gt ()
    "Same as `c-electric-lt', but for \">\"."
    (interactive)
    (when electric-pair-mode
      (if (and (save-excursion
		 (beginning-of-line)
		 (equal (char-after) ?#))
	       (equal (char-after) ?>))
	  (forward-char 1)
	(insert ?>))))
  (keymap-set c-mode-map "<" 'tr/c-electric-lt)
  (keymap-set c-mode-map ">" 'tr/c-electric-gt)
  (keymap-set c-mode-map "C-d" 'tr/kill-forward-char)
  ;; (add-hook 'c-mode-hook 'tree-sitter-hl-mode)
  (modify-syntax-entry ?_ "w" c-mode-syntax-table)
  :custom
  (c-default-style "linux")
  (c-indent-level 8)
  (c-indent-tabs-mode t)
  (tab-width 8))

(use-package gdb-mi
  :custom
  (gdb-many-windows 1))

(use-package tex-mode
  :config
  (add-to-list 'auto-mode-alist '("\\.lill\\'" . latex-mode))
  (defun tr/tex-insert-dollar ()
    "Custom insertion of the \\=`$\\=' in `latex-mode'.

If the character under the point is \\=`$\\=', then the
point just moves forward.  If it is something else, the
regular `self-insert-command' is called."
    (interactive)
    (if (equal (char-after) ?$)
	(forward-char)
      (self-insert-command 1 ?$)))
  (defun tr/sentence-end-double-space-nil ()
    "Set `sentence-end-double-space' to nil locally."
    (setq-local sentence-end-double-space nil))
  (keymap-set latex-mode-map "$" 'tr/tex-insert-dollar)
  (add-hook 'latex-mode-hook 'display-line-numbers-mode)
  (add-hook 'latex-mode-hook 'tr/sentence-end-double-space-nil))

(use-package man
  ;; TODO: move parts to core
  :custom
  (Man-notify-method 'aggressive)
  (Man-width 80)
  :config
  (defun tr/man-copy-c-include ()
    "Copy the include statement at point."
    (interactive)
    (save-excursion
      (end-of-line)
      (let ((boundary (point)))
	(beginning-of-line)
	(search-forward "#" boundary)
	(copy-region-as-kill (match-beginning 0) boundary))))
  (keymap-set Man-mode-map "C-c M-w" 'tr/man-copy-c-include)
  :bind
  ("C-c m" . man))

(use-package compile
  :custom
  (compilation-always-kill t)
  (compilation-scroll-output 'first-error))

(use-package project
  :config
  (defun tr/project-find (file)
    "Find FILE in current project.

It is created if it does not exist."
    (find-file (concat (project-root (project-current)) file)))

  (defun project-gitignore ()
    "Open current project's .gitignore file."
    (interactive)
    (tr/project-find ".gitignore"))

  (keymap-global-set "C-c n i" 'project-gitignore)

  (defun project-makefile ()
    "Open current project's Makefile."
    (interactive)
    (tr/project-find "Makefile"))

  (keymap-global-set "C-c n m" 'project-makefile)

  (defun project-license ()
    "Open current project's license."
    (interactive)
    (tr/project-find "LICENSE"))

  (keymap-global-set "C-c n l" 'project-license)

  (defun project-readme ()
    "Open current project's README.

The supported extensions are :
- \".md\",
- \".org\",
- none."
    (interactive)
    (let ((root (project-root (project-current)))
	  (final-message "README not found"))
      (dolist (ext '(".md" ".org" "") (message final-message))
	(let ((current (expand-file-name (concat "README" ext) root)))
	  (when (file-exists-p current)
	    (setq final-message "")
	    (find-file current))))))

  (keymap-global-set "C-c n r" 'project-readme)

  (defun project-clear-buffers ()
    "Kill all current project buffers and start Dired."
    (interactive)
    (let ((root (project-root (project-current))))
      (project-kill-buffers t) ;; no prompt
      (dired root)
      (delete-other-windows)))

  (keymap-global-set "C-x p K" 'project-clear-buffers)

  (defun project-xref-regen (project)
    "Regenerate the TAGS file for current project."
    (interactive (list (project-current)))
    (let ((default-directory (project-root project)))
      (shell-command "find . -type f -name \"*.[chly]\" -o -name \"Makefile\" \
                                -o -name \"*.tex\" -o -name \"*.py\" | etags -")
      (visit-tags-table "TAGS"))))

(use-package inf-lisp
  :custom
  (inferior-lisp-program "sbcl"))

(use-package scheme
  :custom
  (scheme-program-name "guile")
  :config
  (add-hook 'scheme-mode-hook 'enable-paredit-mode)
  (add-hook 'scheme-mode-hook 'turn-on-prettify-symbols-mode))

(use-package diff-mode
  :config
  (set-face-attribute 'diff-indicator-added nil
		      :foreground "#4c4f69"
		      :background "#40a02b")
  (set-face-attribute 'diff-indicator-removed nil
		      :foreground "#4c4f69"
		      :background "#d20f39")
  (set-face-attribute 'diff-refine-added nil
		      :foreground "#4c4f69"
		      :background "#40a02b")
  (set-face-attribute 'diff-refine-removed nil
		      :foreground "#4c4f69"
		      :background "#d20f39")
  (set-face-attribute 'diff-added nil
		      :foreground "#4c4f69"
		      :background "#40a02b")
  (set-face-attribute 'diff-removed nil
		      :foreground "#4c4f69"
		      :background "#d20f39"))

;;;; Third-Party Packages
(require 'use-package-ensure)
(setq use-package-always-ensure t)

(use-package vertico
  :defer nil
  :custom
  (vertico-count 6)
  (vertico-cylce t)
  :bind
  (:map vertico-map
	("C-n" . vertico-next)
	("C-p" . vertico-previous))
  :config
  (vertico-mode 1))

(use-package orderless
  :defer nil
  :custom
  (completion-styles '(orderless basic))
  (completion-category-overrides '((file (styles basic partial-completion)))))

(use-package corfu
  :defer nil
  :custom
  (corfu-cycle t)
  :bind
  (:map corfu-map
	("C-n" . corfu-next)
	("C-p" . corfu-previous))
  :config
  (global-corfu-mode 1))

(use-package marginalia
  :defer nil
  :config
  (marginalia-mode 1))

(use-package yasnippet
  :defer nil
  :config
  (setq yas-snippet-dirs (list (expand-file-name "snippets/"
						 user-emacs-directory)))
  (yas-global-mode 1))

(use-package consult
  :defer nil
  :bind
  ("C-x b"   . consult-buffer)
  ("C-c b"   . consult-bookmark)
  ("C-x p b" . consult-project-buffer))

(use-package cape
  :defer nil
  :config
  (add-to-list 'completion-at-point-functions 'cape-file)
  (add-to-list 'completion-at-point-functions 'cape-keyword)
  (add-to-list 'completion-at-point-functions 'cape-dabbrev))

(use-package catppuccin-theme
  :defer nil
  :config
  (setq catppuccin-flavor 'latte)
  (load-theme 'catppuccin t))

(use-package solarized-theme)

(use-package beacon
  :defer nil
  :custom
  (beacon-dont-blink-major-modes '(t magit-status-mode))
  :config
  (beacon-mode 1)
  (keymap-global-set "C-c SPC" 'beacon-blink)
  (keymap-global-set "C-c C-SPC" 'beacon-blink))

(use-package all-the-icons
  :defer nil)

(use-package doom-modeline
  :defer nil
  :custom
  (doom-modeline-buffer-encoding nil)
  (doom-modeline-icon t)
  (doom-modeline-buffer-modification-icon t)
  (doom-modeline-buffer-state-icon t)
  :config
  (doom-modeline-mode 1))

(use-package rainbow-delimiters
  :defer nil
  :config
  (add-hook 'find-file-hook 'rainbow-delimiters-mode))

(use-package rainbow-mode)

(use-package markdown-mode)

(use-package hl-todo
  :defer nil
  :custom
  (hl-todo-keyword-faces '(("TODO"  . tr/hl-todo-todo-face)
			   ("IDEA"  . tr/hl-todo-idea-face)
			   ("FIXME" . tr/hl-todo-fixme-face)))
  :config
  (defun tr/hl-todo-setup ()
    "Set up `hl-todo'."
    (set-face-attribute 'hl-todo nil
			:foreground "#eff1f5"
			:weight 'ultra-bold))
  (defface tr/hl-todo-todo-face
    '((t (:inherit hl-todo
	  :foreground "#eff1f5"
	  :background "#8839ef")))
    "Face for TODOs.")
    (defface tr/hl-todo-idea-face
    '((t (:inherit hl-todo
	  :foreground "#eff1f5"
	  :background "#40a02b")))
    "Face for IDEAs.")
  (defface tr/hl-todo-fixme-face
    '((t (:inherit hl-todo
	  :foreground "#eff1f5"
	  :background "#d20f39")))
    "Face for FIXMEs.")
  ;; (tr/hl-todo-setup)
  (global-hl-todo-mode))

(use-package magit)

(use-package magit-todos
  :defer nil
  :config
  (magit-todos-mode 1))

(use-package tuareg
  :config
  (defun tr/ocaml-prettify-symbols ()
    (setq-local prettify-symbols-alist
		tuareg-prettify-symbols-basic-alist)
    (prettify-symbols-mode 1))
  (add-hook 'tuareg-mode-hook 'tr/ocaml-prettify-symbols)
  (add-hook 'tuareg-interactive-mode-hook 'tr/ocaml-prettify-symbols)
  (defun tr/ocaml-next-problem ()
    "Jump to next OCaml problem."
    (interactive)
    (if (eq major-mode 'tuareg-mode)
	(let ((problem (string-to-number (file-name-base buffer-file-name))))
	  (if (= problem 99)
	      (message "All problems completed !")
	    (find-file (format "%02d.ml" (1+ problem)))))
      (message "error: not editing OCaml code"))))

(use-package sly)

(use-package yaml-mode)

(use-package gnuplot-mode)

(use-package multiple-cursors
  :defer nil)

(use-package hydra
  :defer nil
  :config
  (defhydra hydra-zoom (global-map "C-c z")
    "zoom"
    ("p" text-scale-increase)
    ("+" text-scale-increase)
    ("n" text-scale-decrease)
    ("-" text-scale-decrease)))

(use-package package-build
  :init
  (defcustom tr/package-build-dir
    (expand-file-name "package-build/" user-emacs-directory)
    "Main directory for `package-build' stuff.")
  (defun tr/publish-elisp-packages ()
    "Publish my Emacs packages upstream."
    (interactive)
    (async-shell-command "~/.dotfiles/publish-elisp-packages.sh"))
  :custom
  (package-build-working-dir (expand-file-name "working/" tr/package-build-dir))
  (package-build-archive-dir (expand-file-name "packages/" tr/package-build-dir))
  (package-build-recipes-dir (expand-file-name "recipes/" tr/package-build-dir)))

(use-package paredit)

(use-package clang-format)

(use-package shader-mode
  :mode "\\.vert\\'"
  :mode "\\.frag\\'")

;;;; Final Step
(defun tr/set-faces ()
  (set-face-attribute 'default nil
		      :font "Mononoki Nerd Font"
		      :height 128)
  (set-face-attribute 'fixed-pitch nil
		      :font "Mononoki Nerd Font"
		      :height 128)
  (set-face-attribute 'variable-pitch nil
		      :font "Free Sans"
		      :height 128)
  (set-face-attribute 'help-key-binding nil
		      :foreground "blue")
  (tr/hl-todo-setup)
  )

(defun tr/emacs-client-frame-setup ()
  "Setup for Emacsclient frames."
  (add-hook 'after-make-frame-functions
	    (lambda (frame)
	      (with-selected-frame frame
		(tr/set-faces)
		(setq doom-modeline-icon t)))))

(defun tr/emacs-frame-setup ()
  "Setup for non-client Emacs frames."
  (tr/set-faces))

(if (daemonp)
    (tr/emacs-client-frame-setup)
  (tr/emacs-frame-setup))

;;;; Mail
(let ((mail-config-file (tr/expand-from-dotfiles-dir "emacs/mail.el")))
  (when (file-exists-p mail-config-file)
    (load mail-config-file)))

;; Local Variables:
;; eval: (when (functionp 'rainbow-mode) (rainbow-mode 1))
;; End:
